from abstar import run, run_standalone, main, parse_arguments, validate_args
from preprocess import fastqc, adapter_trim, quality_trim

from pkg_resources import get_distribution, DistributionNotFound
import os.path

try:
    _dist = get_distribution('abstar')
    # Normalize case for Windows systems
    dist_loc = os.path.normcase(_dist.location)
    here = os.path.normcase(__file__)
    if not here.startswith(os.path.join(dist_loc, 'abstar')):
        # not installed, but there is another version that *is*
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = 'Please install AbStar before checking the version'
else:
    __version__ = _dist.version



#!/usr/bin/python
# filename: regions.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import re
import sys
import traceback

from Bio.Seq import Seq
from Bio.Alphabet import generic_dna

from abtools import log


def regions(blast_result):
    '''
    Returns a VarRegions (or JoinRegions) object, containing a variety of information about
    Variable (or Joining) gene regions.  Both Regions objects contain equivalent attributes
    to ease downstream handling.

    Input is a BlastResult object for a variable gene.
    '''
    global logger
    logger = log.get_logger(__name__)
    try:
        if blast_result.gene_type == 'variable':
            return VarRegions(blast_result)
        if blast_result.gene_type == 'joining':
            return JoinRegions(blast_result)
    except Exception:
        logger.debug('REGIONS ERROR: {} {}'.format(blast_result.id, blast_result.seq))
        logger.debug(traceback.format_exc())



class VarRegions(object):
    '''
    Structure for information about antibody Variable gene regions.

    Input is a BlastResult object for a variable gene.
    '''
    def __init__(self, blast_result):
        self.raw_positions = self._get_region_positions(blast_result)
        self.adjusted_positions = self._adjusted_region_positions(blast_result)
        self.nt_seqs = self._get_region_nt_sequences(blast_result)
        self.nt_lengths = self._get_region_lengths(self.nt_seqs)
        self.aa_seqs = self._translate_regions(blast_result)
        self.aa_lengths = self._get_region_lengths(self.aa_seqs)
        self.germline_nt_seqs = self._get_region_nt_sequences(blast_result, germline=True)
        self.germline_aa_seqs = self._translate_regions(blast_result, germline=True)


    def _get_region_positions(self, br):
        '''
        Parses the region positions file to get the region positions for a
        specific germline gene.

        Input is a BlastResult object.

        Output is a list of region end positions, with the end position being
        'None' if the region is not present in the sequence (for example, if the
        sequence is truncated and doesn't contain FR1)
        '''
        region_positions = []
        mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        regions_file = os.path.join(mod_dir, 'utils/germline_data/v_region_info_{}'.format(br.species.lower()))
        with open(regions_file) as f:
            for line in f:
                if line.startswith('#'):
                    continue
                sline = line.split()
                if sline[0] == br.top_germline:
                    region_positions = [int(i) for i in sline[1:]]
                    break
        if br.insertions:
            region_positions = self._recalibrate_region_positions(br, region_positions)
        return region_positions


    def _recalibrate_region_positions(self, br, positions):
        '''
        Recalibrates the region position numbering to account for insertions.

        Input is a BlastResult object and a list of region positions.

        Output is a list of recalibrated region end positions.
        '''
        if br.insertions:
            insertions = [(i['pos'], i['len']) for i in br.insertions]
            insertions.sort(key=lambda x: x[0])
            for i in insertions:
                istart = i[0]
                ilength = i[1]
                adjustments = []
                for p in positions:
                    adjustments.append(0 if p <= istart else ilength)
                positions = [p + a for p, a in zip(positions, adjustments)]
        return positions


    def _adjusted_region_positions(self, br):
        '''
        Adjusts the region positions relative to the start of the query sequence.
        If the query sequence is sufficiently truncated that a region is not present,
        the end position for that region will be set to 'None'.

        Input is a BlastResult object.

        Output is a list of adjusted region end positions.
        '''
        adjusted_positions = [p - br.germline_start for p in self.raw_positions]
        return [p if p > 0 else None for p in adjusted_positions]


    def _get_region_nt_sequences(self, br, germline=False):
        '''
        Identifies the nucleotide sequence for each region of a variable gene.

        Input is a BlastResult object and, optionally, the germline flag. If set
        to True, region sequences for the top germline match are returned.

        Output is a dict of region sequences.
        '''
        # regions = self._regions(br)
        regions = self._regions()
        region_nt_seqs = {}
        start = 0
        if germline:
            alignment = br.germline_alignment
        else:
            alignment = br.query_alignment
        for r in regions:
            reg = r[0]
            end = r[1]
            if end is not None:
                region_nt_seqs[reg] = alignment[start:end]
                if end > len(alignment) and reg == 'FR3':
                    logger.debug('TRUNCATED FR3: {}'.format(br.id))
                    logger.debug('OLD: {}'.format(region_nt_seqs['FR3']))
                    region_nt_seqs[reg] = self._fix_truncated_fr3_alignment(br, start, end, germline)
                    logger.debug('NEW: {}'.format(region_nt_seqs['FR3']))
                start = end
            else:
                region_nt_seqs[reg] = None
        return region_nt_seqs


    def _region_list(self):
        return ['FR1', 'CDR1', 'FR2', 'CDR2', 'FR3']


    def _get_region_lengths(self, regions):
        lengths = {}
        for r in regions:
            if regions[r]:
                lengths[r] = len(regions[r])
            else:
                lengths[r] = 0
        return lengths


    # def _regions(self, region_positions):
    def _regions(self):
        '''
        Produces a list of regions and region end positions for all regions.

        Input is a list of region positions.

        Output is a list of tuples, of the format:
        (region_name, region_end_position)
        '''
        all_regions = self._region_list()
        return zip(all_regions, self.adjusted_positions)


    def _translate_regions(self, br, germline=False):
        '''
        Translates the region sequence.

        Input is a BlastResult object and, optionally, the germline flag. If germline is set
        to True, translated regions for the top germline match will be returned.

        Output is a dict of region sequences.
        '''
        if germline:
            nt_seqs = self.germline_nt_seqs
        else:
            nt_seqs = self.nt_seqs
        aa_seqs = {}
        rf = br.germline_start % 3
        rf_offset = (rf * 2) % 3
        first_region = True
        regions = self._region_list()
        for i, r in enumerate(regions):
            if r in nt_seqs.keys():
                if not nt_seqs[r]:
                    aa_seqs[r] = None
                    continue
                if i <= len(regions) - 2:
                    nt_seqs = self._fix_region_spanning_indel(nt_seqs, r, regions[i + 1:])
                if first_region:
                    aa_seqs[r] = str(Seq(nt_seqs[r][rf_offset:].replace('-', ''), generic_dna).translate())
                    if '-' in nt_seqs[r]:
                        aa_seqs[r] = self._fix_translated_indels(nt_seqs[r], aa_seqs[r])
                    first_region = False
                else:
                    trim = len(nt_seqs[r].replace('-', '')) % 3
                    seq_to_trans = nt_seqs[r].replace('-', '')[:-trim] if trim else nt_seqs[r].replace('-', '')
                    aa_seqs[r] = str(Seq(seq_to_trans, generic_dna).translate())
                    if '-' in nt_seqs[r]:
                        aa_seqs[r] = self._fix_translated_indels(nt_seqs[r], aa_seqs[r])
            else:
                aa_seqs[r] = None
        if germline:
            self.germline_nt_seqs = nt_seqs
        else:
            self.nt_seqs = nt_seqs
        return aa_seqs


    def _fix_region_spanning_indel(self, nt_seqs, r1, next_regions):
        r1_len = len(nt_seqs[r1].rstrip('-'))
        if r1_len % 3 == 0:
            return nt_seqs
        logger.debug(next_regions)
        reg_iter = iter(next_regions)
        r2 = reg_iter.next()
        while len(nt_seqs[r2].strip('-')) == 0:
            logger.debug(r2)
            r2 = reg_iter.next()
        r2_len = len(nt_seqs[r2].lstrip('-'))
        if r1_len % 3 > r2_len % 3:
            len_to_move = r2_len % 3
            chunk_to_move = nt_seqs[r2].lstrip('-')[:len_to_move]
            r1_gap = len(nt_seqs[r1]) - r1_len - len_to_move
            r2_gap = len(nt_seqs[r2]) - r2_len + len_to_move
            nt_seqs[r1] = '{}{}{}'.format(nt_seqs[r1].rstrip('-'), chunk_to_move, '-' * r1_gap)
            nt_seqs[r2] = '{}{}'.format('-' * r2_gap, nt_seqs[r2].lstrip('-')[len_to_move:])
        else:
            len_to_move = r1_len % 3
            chunk_to_move = nt_seqs[r1].rstrip('-')[-len_to_move:]
            r1_gap = len(nt_seqs[r1]) - r1_len + len_to_move
            r2_gap = len(nt_seqs[r2]) - r2_len - len_to_move
            nt_seqs[r1] = '{}{}'.format(nt_seqs[r1].rstrip('-')[:-len_to_move], '-' * r1_gap)
            nt_seqs[r2] = '{}{}{}'.format('-' * r2_gap, chunk_to_move, nt_seqs[r2].lstrip('-'))
        return nt_seqs


    def _fix_translated_indels(self, nt_seq, aa_seq):
        for indel in re.finditer('-+', nt_seq):
            s = indel.start() / 3
            e = indel.end() / 3
            i = '-' * (e - s)
            return aa_seq[:s] + i + aa_seq[s:]


    def _fix_truncated_fr3_alignment(self, br, start, end, germline):
        if germline:
            alignment = br.germline_alignment
            seq = br.germline_seq[br.germline_end + 1:]
        else:
            alignment = br.query_alignment
            seq = br.input_sequence[br.query_end + 1:]
        trunc = end - len(alignment)
        return alignment[start:] + seq[:trunc]


class JoinRegions(object):
    '''
    Structure for information about antibody Joining gene regions.

    Input is a BlastResult object.
    '''
    def __init__(self, br):
        self.fix_v_overlap = False
        self.v_overlap_length = 0
        start = self._get_fr4_nt_start(br)
        end = self._get_fr4_nt_end(br)
        self.raw_positions = [start, end]
        self.adjusted_positions = self.raw_positions
        self.nt_seqs = self._get_fr4_nt_seq(br, start, end)
        self.nt_lengths = self._get_region_lengths(self.nt_seqs)
        self.aa_seqs = self._translate_regions(self.nt_seqs)
        self.aa_lengths = self._get_region_lengths(self.aa_seqs)
        self.germline_nt_seqs = self._get_fr4_germ_nt_seq(br, start, end)
        self.germline_aa_seqs = self._translate_regions(self.germline_nt_seqs)


    def _get_fr4_nt_start(self, br):
        '''
        Returns the start position (in the joining gene alignment) of FR4.

        Input is a BlastResult object.
        '''
        mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        regions_file = os.path.join(mod_dir, 'utils/germline_data/j_region_info_{}'.format(br.species.lower()))
        with open(regions_file) as f:
            for line in f:
                if line.startswith('#'):
                    continue
                sline = line.split()
                if sline[0] == br.top_germline:
                    raw_fr4_start = int(sline[2])
                    break
        # in some rare cases where the V-gene alignment overlaps
        # the J-gene alignment, which results in the first nt or two
        # of FR4 being truncated (because they're part of the V-gene alignment)
        if br.chain != 'heavy' and raw_fr4_start - br.germline_start < 0:
            self.fix_v_overlap = True
            self.v_overlap_length = abs(raw_fr4_start - br.germline_start)
            logger.debug('CHAIN: {}'.format(br.chain))
            logger.debug('V-OVERLAP FOUND: {}'.format(br.id))
            logger.debug('V-OVERLAP LENGTH: {}'.format(self.v_overlap_length))
        return max(0, raw_fr4_start - br.germline_start)


    def _get_fr4_nt_end(self, br):
        '''
        Returns the end position (in the joining gene alignment) of FR4.

        Input is a BlastResult object.
        '''
        return br.query_end


    def _get_fr4_nt_seq(self, br, start, end):
        '''
        Returns the nucleotide sequence of FR4.

        Input is a BlastResult object and the start and end positions of FR4.
        '''
        return {'FR4': br.query_alignment[start:end]}


    def _get_fr4_germ_nt_seq(self, br, start, end):
        '''
        Returns the germline nucleotide sequence of FR4.

        Input is a BlastResult object and the start and end positions of FR4.
        '''
        return {'FR4': br.germline_alignment[start:end]}


    def _translate_regions(self, regions):
        '''
        Returns the translated sequence for each region.

        Input is a dict of nucleotide sequences for each region.
        '''
        translated = {}
        for r in regions:

            trim = len(regions[r].replace('-', '')) % 3
            seq_for_trans = regions[r].replace('-', '')[:-trim] if trim else regions[r].replace('-', '')
            translated[r] = str(Seq(seq_for_trans, generic_dna).translate())
            if '-' in regions[r]:
                translated[r] = self._fix_translated_indels(regions[r], translated[r])
        return translated


    def _fix_translated_indels(self, nt_seq, aa_seq):
        for indel in re.finditer('-+', nt_seq):
            s = indel.start() / 3
            e = indel.end() / 3
            i = '-' * (e - s)
            return aa_seq[:s] + i + aa_seq[s:]


    def _get_region_lengths(self, regions):
        '''
        Returns the sequence lengths of each region.

        Input is a dict of nucleotide or amino acid sequences for each region.
        '''
        lengths = {}
        for r in regions:
            if regions[r]:
                lengths[r] = len(regions[r])
            else:
                lengths[r] = 0
        return lengths



#!/usr/bin/python
# filename: vdj.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import logging
import os
import sys
import tempfile
import traceback

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna

from celery.utils.log import get_task_logger

from abstar.utils import blast
from abstar.utils.queue.celery import celery

from abtools import log
from abtools.alignment import local_alignment
from abtools.sequence import Sequence


class VDJ(object):
    '''
    Data structure for managing germline gene assignments for a single antibody sequence.

    Input is a Sequence object, V and J BlastResult objects, and (optionally) a
    DiversityResult object.

    VDJ objects without an identified V or J gene will have 'None' for the unidentified gene
    segment and the 'rearrangement' attribute will be False.
    '''

    def __init__(self, seq, args, v, j, d=None):
        super(VDJ, self).__init__()
        self.id = seq.id
        self.raw_input = seq.sequence
        self.raw_query = seq._input_sequence
        self.args = args
        self.strand = v.strand
        self._uaid = None
        self._isotype = None
        self.v = v
        self.j = j
        self.d = d
        if not self._valid_assignments():
            self.rearrangement = False
        elif all([v is not None, j is not None]):
            try:
                self.rearrangement = True
                logger.debug('V-BLAST INPUT: {}'.format(v.seq.sequence))
                logger.debug('V-BLAST QUERY: {}'.format(v.query_alignment))
                logger.debug('J-BLAST INPUT: {}'.format(j.seq.sequence))
                logger.debug('J-BLAST QUERY: {}'.format(j.query_alignment))
                self._get_attributes()
            except:
                self.rearrangement = False
                logger.debug('VDJ ATTRIBUTE ERROR: {}'.format(seq.id))
                logger.debug(traceback.format_exc())
        else:
            if v is None:
                logger.debug('V ASSIGNMENT ERROR: {}'.format(seq.id))
            if j is None:
                logger.debug('J ASSIGNMENT ERROR: {}'.format(seq.id))
            self.rearrangement = False

    @property
    def uaid(self):
        if self._uaid is None:
            if self.args.uaid == 0:
                self._uaid = ''
            elif self.args.uaid > 0:
                self._uaid = self.raw_query[:self.args.uaid]
            elif self.args.uaid < 0:
                self._uaid = self.raw_query[self.args.uaid:]
            else:
                self._uaid = ''
        return self._uaid

    @property
    def isotype(self):
        return self._isotype


    def _valid_assignments(self):
        if any([self.v is None, self.j is None]):
            return False
        if self.v.chain != self.j.chain:
            logger.debug("ASSIGNMENT ERROR: {} - V and J chains don't match".format(self.id))
            return False
        return True

    def _get_attributes(self):
        'Adds VDJ attributes, only for VDJ objects with an identified V and J gene.'
        self.species = self.v.species
        self.chain = self._get_chain()
        self.query_reading_frame = self._query_reading_frame()
        self.v_start = self._get_v_start()
        self.v_end = self._get_v_end()
        self.j_start = self._get_j_start()
        self.j_end = self._get_j_end()
        if self.chain in ['kappa', 'lambda'] and self.j.regions.fix_v_overlap:
            self._fix_v_overlap()
        if self.d:
            self.d_start = self._get_d_start()
            self.d_end = self._get_d_end()
            self.n1_start = self._get_n1_start()
            self.n1_end = self._get_n1_end()
            self.n2_start = self._get_n2_start()
            self.n2_end = self._get_n2_end()
            self.n_start = None
            self.n_end = None
        else:
            self.d_start = None
            self.d_end = None
            self.n1_start = None
            self.n1_end = None
            self.n2_start = None
            self.n2_end = None
            self.n_start = self._get_n_start()
            self.n_end = self._get_n_end()

        self.gapped_vdj_nt = self._gapped_vdj_nt()
        self.vdj_nt = self._vdj_nt()
        self.gapped_vdj_region_string = self._get_gapped_vdj_region_string()
        self.vdj_region_string = self._get_vdj_region_string()
        self.vdj_aa = self._vdj_aa()
        self.junction = self._get_junction()
        self.gapped_vdj_germ_nt = self._gapped_vdj_germ_nt()
        self.vdj_germ_nt = self.gapped_vdj_germ_nt.replace('-', '')
        self.vdj_germ_aa = self._vdj_germ_aa()
        self.germ_junction = self._get_junction(germ=True)
        self.codons = self._get_codons()
        self.gapped_codons = self._get_codons(gapped=True)
        self.productive = self._check_productivity()
        if self.args.isotype:
            from abstar.utils.isotype import get_isotype
            self._isotype = get_isotype(self)
        else:
            self._isotype = ''
        if not self.junction:
            self.rearrangement = False

    def _get_chain(self):
        'Returns the antibody chain.'
        try:
            if self.v.top_germline.startswith('IGH'):
                return 'heavy'
            elif self.v.top_germline.startswith('IGK'):
                return 'kappa'
            elif self.v.top_germline.startswith('IGL'):
                return 'lambda'
        except:
            logger.debug('GET CHAIN ERROR: {}, {}'.format(self.id,
                                                          self.v.top_germline))
            return ''

    def _query_reading_frame(self):
        'Returns the reading frame of the query sequence.'
        try:
            return self.v.germline_start % 3
        except:
            logger.debug('QUERY READING FRAME ERROR: {}, {}'.format(self.id,
                                                                    self.v.germline_start))

    def _fix_v_overlap(self):
        logger.debug('FIXING V_OVERLAP: {}'.format(self.id))
        nt = self.j.regions.v_overlap_length
        # fix the query sequence
        new_fr4_nt = self.v.query_alignment[-nt:] + self.j.regions.nt_seqs['FR4']
        self.j.regions.nt_seqs['FR4'] = new_fr4_nt
        self.j.regions.aa_seqs['FR4'] = str(Seq(new_fr4_nt, generic_dna).translate())
        # fix the germline sequence
        new_fr4_germ_nt = self.v.germline_alignment[-nt:] + self.j.regions.germline_nt_seqs['FR4']
        self.j.regions.germline_nt_seqs['FR4'] = new_fr4_germ_nt
        self.j.regions.germline_aa_seqs['FR4'] = str(Seq(new_fr4_germ_nt, generic_dna).translate())

    def _gapped_vdj_nt(self):
        try:
            gapped_vdj_nt = self.v.query_alignment + \
                self.j.input_sequence[:self.j.query_start] + \
                self.j.query_alignment
            return gapped_vdj_nt
        except:
            logger.debug('VDJ NT ERROR: {}, {}'.format(self.id, self.raw_query))

    def _vdj_nt(self):
        'Returns the nucleotide sequence of the VDJ region.'
        return self.gapped_vdj_nt.replace('-', '')

    def _vdj_aa(self):
        'Returns the amino acid sequence of the VDJ region.'
        offset = (self.query_reading_frame * 2) % 3
        trim = len(self.vdj_nt) - (len(self.vdj_nt[offset:]) % 3)
        translated_seq = Seq(self.vdj_nt[offset:trim], generic_dna).translate()
        return str(translated_seq)

    def _gapped_vdj_germ_nt(self):
        'Returns the nucleotide sequence of the VDJ region.'
        try:
            if self.d:
                germ_junction = self.junction.n1_nt + \
                    self.d.germline_alignment + \
                    self.junction.n2_nt
            else:
                germ_junction = self.junction.n_nt
            germ_vdj_nt = self.v.germline_alignment + \
                germ_junction + self.j.germline_alignment
            return germ_vdj_nt
        except:
            logger.debug('VDJ GERM NT ERROR: {}, {}'.format(self.id, self.raw_query))
            logger.debug(traceback.format_exc())

    def _vdj_germ_aa(self):
        'Returns the amino acid sequence of the VDJ region.'
        offset = (self.query_reading_frame * 2) % 3
        trim = len(self.vdj_germ_nt) - (len(self.vdj_germ_nt[offset:]) % 3)
        translated_seq = Seq(self.vdj_germ_nt[offset:trim], generic_dna).translate()
        return str(translated_seq)

    def _get_junction(self, germ=False):
        from abstar.utils import junction
        return junction.get_junction(self, germ=germ)

    def _get_v_start(self):
        return self.v.query_start

    def _get_v_end(self):
        return len(self.v.query_alignment)

    def _get_j_start(self):
        return self.v_end + self.j.query_start

    def _get_j_end(self):
        return self.j_start + len(self.j.query_alignment)

    def _get_n1_start(self):
        return self.v_end

    def _get_n1_end(self):
        return self.d_start

    def _get_d_start(self):
        return self.v_end + self.d.query_start

    def _get_d_end(self):
        return self.d_start + len(self.d.query_alignment)

    def _get_n2_start(self):
        return self.d_end

    def _get_n2_end(self):
        return self.j_start

    def _get_n_start(self):
        return self.v_end

    def _get_n_end(self):
        return self.j_start

    def _get_vdj_region_string(self):
        rstring = ''
        for s, r in zip(self.gapped_vdj_nt, self.gapped_vdj_region_string):
            if s == '-':
                continue
            rstring += r
        return rstring

    def _get_gapped_vdj_region_string(self):
        region_string = ''
        region_string += 'V' * self.v_end
        if self.d:
            region_string += 'N' * (self.d_start - self.v_end)
            region_string += 'D' * (self.d_end - self.d_start)
            region_string += 'N' * (self.j_start - self.d_end)
        else:
            region_string += 'N' * (self.j_start - self.v_end)
        region_string += 'J' * (self.j_end - self.j_start)
        return region_string

    def _get_codons(self, gapped=False):
        from abstar.utils import codons
        return codons.parse_codons(self, gapped=gapped)

    def _check_productivity(self):
        from abstar.utils import productivity
        return productivity.check_productivity(self)

    def _build_output(self, output_type):
        from abstar.utils import output
        return output.build_output(self, output_type)


@celery.task
def run(seq_file, output_dir, args):
    '''
    Wrapper function to multiprocess (or not) the assignment of V, D and J
    germline genes. Also writes the JSON-formatted output to file.

    Input is a a FASTA-formatted file of antibody sequences and the output directory.
    Optional input items include the species (supported species: 'human'); length of
    the unique antibody identifier (UAID); and debug mode (which forces single-threading
    and prints more verbose errors.)

    Output is the number of functional antibody sequences identified in the input file.
    '''
    try:
        global logger
        if args.cluster:
            logger = get_task_logger(__name__)
        else:
            logger = log.get_logger(__name__)
        output_filename = os.path.basename(seq_file)
        if args.output_type == 'json':
            output_file = os.path.join(output_dir, output_filename + '.json')
        elif args.output_type in ['imgt', 'hadoop']:
            output_file = os.path.join(output_dir, output_filename + '.txt')
        vdj_output = process_sequence_file(seq_file, args)
        if not vdj_output:
            return None
        clean_vdjs = [vdj for vdj in vdj_output if vdj.rearrangement]
        output_count = write_output(clean_vdjs, output_file, args.output_type, args.pretty, args.padding)
        return (output_file, output_count)
    except:
        logger.debug(traceback.format_exc())
        raise Exception("".join(traceback.format_exception(*sys.exc_info())))
        # run.retry(exc=exc, countdown=5)


def write_output(output, outfile, output_type, pretty, padding):
    from abstar.utils.output import build_output
    logger.debug("Padding - {}\t Pretty - {}\t".format(padding, pretty))
    output_data = build_output(output, output_type, pretty, padding)
    open(outfile, 'w').write('\n'.join(output_data))
    if output_type in ['json', 'hadoop']:
        return len(output_data)
    else:
        return len(output_data) - 1


def process_sequences(sequences, args):
    global logger
    logger = logging.getLogger()
    seq_file = tempfile.NamedTemporaryFile(delete=False)
    seq_file.write('\n'.join([s.fasta for s in sequences]))
    seq_file.close()
    vdjs = process_sequence_file(seq_file.name, args)
    os.unlink(seq_file.name)
    return vdjs


def process_sequence_file(seq_file, args):
    '''
    Runs BLASTn to identify germline V, D, and J genes.

    Input is a Sequence object.

    Output is a list of VDJ objects.
    '''

    # Variable gene assignment
    vs = []
    seqs = [Sequence(s) for s in SeqIO.parse(open(seq_file, 'r'), 'fasta')]
    v_blast_records = blast.blast(seq_file, args.species, 'V')
    for seq, vbr in zip(seqs, v_blast_records):
        try:
            v = assign_germline(seq, vbr, args.species, 'V')
            if v.strand == 'minus':
                seq.sequence = seq.reverse_complement
            logger.debug('ASSIGNED V-GENE: {}, {}'.format(seq.id, v.top_germline))
        except Exception:
            logger.debug('V-GENE ASSIGNMENT ERROR: {}'.format(seq.id))
            logger.debug(traceback.format_exc())
            v = None
        finally:
            if v:
                # v_end = v.query_end + v.query_start + 1
                v_end = v.query_end + 1
                if len(seq.sequence[v_end:]) <= 10:
                    logger.debug('REMANING REGION TOO SHORT AFTER V-ALIGNMENT REMOVAL: {}'.format(v.id))
                    # v = None
                    logger.debug('PERFORMING SECOND V-GENE REALIGNMENT: {}'.format(v.id))
                    v.realign_variable(v.top_germline,
                        match=30, mismatch=-20,
                        gap_open_penalty=220, gap_extend_penalty=1)
                    v.annotate()
                    # v_end = v.query_end + v.query_start + 1
                    v_end = v.query_end + 1
                    if len(seq.sequence[v_end:]) <= 10:
                        logger.debug('REMANING REGION TOO SHORT AFTER SECOND V-ALIGNMENT REMOVAL: {}'.format(v.id))
                        v = None
            vs.append((seq, v))
    v_blast_results = [v[1] for v in vs if v[1]]
    seqs = [v[0] for v in vs if v[1]]
    failed_seqs = [v[0] for v in vs if not v[1]]
    logger.debug('V-ASSIGNMENT RESULT: for {}, {} of {} sequences failed v-gene assignment'.format(
        os.path.basename(seq_file), len(failed_seqs), len(seqs) + len(failed_seqs)))
    for fs in failed_seqs:
        logger.debug('NO V-GENE ASSIGNMENT: {}'.format(fs.id))
    if not v_blast_results:
        seq_filename = os.path.basename(seq_file)
        logger.debug('NO VALID REARRANGEMENTS IN FILE: {}'.format(seq_filename))
        return None

    # Joining gene assignment
    js = []
    j_blastin = blast.build_j_blast_input(seqs, v_blast_results)
    j_blast_records = blast.blast(j_blastin.name, args.species, 'J')
    j_seqs = [Sequence(s) for s in SeqIO.parse(open(j_blastin.name, 'r'), 'fasta')]
    for i, jdata in enumerate(zip(j_seqs, j_blast_records)):
        j_seq, jbr = jdata
        try:
            j = assign_germline(j_seq, jbr, args.species, 'J')
            if not j:
                logger.debug('NO ASSIGNED J-GENE: {}'.format(j_seq.id))
            logger.debug('ASSIGNED J-GENE: {}, {}'.format(j_seq.id, j.top_germline))
        except:
            logger.debug('J-GENE ASSIGNMENT ERROR: {}'.format(j_seq.id))
            logger.debug(traceback.format_exc())
            try:
                vbr = v_blast_records[i]
                vseq = seqs[i]
                logger.debug('J-GENE ASSIGNMENT ERROR: {}\n{}'.format(j_seq.id,
                                                                          vseq.sequence, ))
                                                                          # vbr.query_alignment))
            except:
                logger.debug('J-GENE ASSIGNMENT ERROR: {}, could not print query info'.format(j_seq.id))
                logger.debug(traceback.format_exc())
            j = None
        finally:
            js.append(j)
    os.unlink(j_blastin.name)

    # Build VDJ objects (including optional D gene assignment)
    vdjs = []
    for seq, v, j in zip(seqs, v_blast_results, js):
        try:
            if not v or not j:
                continue
            if v.chain == 'heavy':
                junc_start = len(v.query_alignment) + v.query_start
                junc_end = junc_start + j.query_start
                junction = Sequence(seq.sequence[junc_start:junc_end], id=seq.id)
                if junction:
                    d = assign_d(junction, args.species)
                    if d is not None:
                        logger.debug('ASSIGNED D-GENE: {}, {}'.format(seq.id, d.top_germline))
                    vdjs.append(VDJ(seq, args, v, j, d))
                    continue
                vdjs.append(VDJ(seq, args, v, j))
            else:
                vdjs.append(VDJ(seq, args, v, j))
        except:
            logger.debug('VDJ ERROR: {}'.format(seq.id))
            logger.debug(traceback.format_exc())
    return vdjs


def assign_germline(seq, blast_record, species, segment):
    '''
    Identifies germline genes for a given antibody sequence (seq).

    Input is a Sequence object, the species of origin, the gene
    segment to be assigned (options are 'V' or 'J') and, optionally, the
    starting and ending points of the possible germline gene location.

    Output is a BlastResult object.
    '''
    blast_result = blast.BlastResult(seq, blast_record, species)
    if blast_result.gene_type == 'variable':
        blast_result.realign_variable(blast_result.top_germline)
    blast_result.annotate()
    return blast_result


def assign_d(seq, species):
    '''
    Identifies the germline diversity gene for a given sequence.
    Alignment is performed using the ssw_wrap.Aligner.align function.

    Input is a junction sequence (as a string) and the species of origin.

    Output is a DiversityResult object.
    '''
    mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    db_file = os.path.join(mod_dir, 'ssw/dbs/{}_D.fasta'.format(species.lower()))
    db_handle = open(db_file, 'r')
    germs = [Sequence(s) for s in SeqIO.parse(db_handle, 'fasta')]
    rc_germs = [Sequence(s.reverse_complement, id=s.id) for s in germs]
    germs.extend(rc_germs)
    db_handle.close()
    alignments = local_alignment(seq, targets=germs,
                                 gap_open_penalty=20, gap_extend_penalty=2)
    alignments.sort(key=lambda x: x.score, reverse=True)
    try:
        return blast.DiversityResult(seq, alignments[:5])
    except IndexError:
        return None



#!/usr/bin/python
# filename: productivity.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import traceback

from abtools import log


def check_productivity(vdj):
    logger = log.get_logger(__name__)
    try:
        problems = 0
        problems += stop_codons(vdj)
        problems += ambig_codons(vdj)
        problems += vdj_agreement(vdj)
        problems += conserved_junc_residues(vdj)
        problems += rearrangement(vdj)
        problems += junction_frame(vdj)
        problems += indels(vdj)
        if problems:
            return 'no'
        return 'yes'
    except:
        logger.debug('PRODUCTIVITY CHECK ERROR: {}'.format(vdj.id))
        logger.debug(traceback.format_exc())


def stop_codons(vdj):
    if '*' in vdj.vdj_aa:
        return 1
    return 0


def ambig_codons(vdj):
    if 'X' in vdj.vdj_aa.upper():
        return 1
    return 0


def vdj_agreement(vdj):
    if vdj.v.top_germline[:3] != vdj.j.top_germline[:3]:
        return 1
    if vdj.d:
        if not vdj.d.top_germline:
            return 0
        if vdj.v.top_germline[:3] != vdj.d.top_germline[:3]:
            return 1
    return 0


def conserved_junc_residues(vdj):
    if not vdj.junction:
        return 1
    start = 'C'
    end = 'W' if vdj.chain == 'heavy' else 'F'
    if vdj.junction.junction_aa[0] != start or vdj.junction.junction_aa[-1] != end:
        return 1
    return 0


def rearrangement(vdj):
    if not vdj.rearrangement:
        return 1
    return 0


def junction_frame(vdj):
    return 0 if vdj.junction.in_frame else 1


def indels(vdj):
	problems = 0
	for indels in [vdj.v.insertions, vdj.j.insertions, vdj.v.deletions, vdj.j.deletions]:
		for i in indels:
			if i['in frame'] == 'no':
				problems += 1
	return problems



#!/usr/bin/python
# filename: isotype.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import logging
import os
import traceback

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna

from abtools import log
from abtools.alignment import local_alignment
from abtools.sequence import Sequence
from abtools.utils.decorators import lazy_property


def get_isotype(vdj):
    logger = log.get_logger(__name__)
    try:
        mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        isotype_file = os.path.join(mod_dir, 'ssw/isotypes/{}_isotypes.fasta'.format(vdj.species))
        isotype_seqs = [Sequence(s) for s in SeqIO.parse(open(isotype_file, 'r'), 'fasta')]
        isotype_seqs += [Sequence((s.id, s.reverse_complement)) for s in isotype_seqs]
        return Isotype(vdj, isotype_seqs)
    except:
        logger.debug('ISOTYPE ERROR: {}\t{}'.format(vdj.id, vdj.raw_query))
        logger.debug(traceback.format_exc())



class Isotype(object):
    """docstring for Isotype"""
    def __init__(self, vdj, isotype_seqs):
        super(Isotype, self).__init__()
        self._alignments = self._get_alignments(vdj, isotype_seqs)
        self.alignment = self._alignments[0]


    @lazy_property
    def isotype(self):
    	if self.alignment.score < 30:
    		return 'unknown'
        return self.alignment.target.id


    @lazy_property
    def score(self):
        return self.alignment.score


    def _get_alignments(self, vdj, isotype_seqs):
        query_region = self._get_isotype_query_region(vdj)
        alignments = local_alignment(query_region, targets=isotype_seqs,
            gap_open_penalty=22, gap_extend_penalty=1)
        return sorted(alignments, key=lambda x: x.score, reverse=True)


    def _get_isotype_query_region(self, vdj):
        aln = local_alignment(vdj.vdj_nt, vdj.raw_input)
        return vdj.raw_input[aln.target_end:]



#!/usr/bin/python
# filename: junction.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import traceback

from Bio.Seq import Seq
from Bio.Alphabet import generic_dna

from abtools import log
from abtools.alignment import global_alignment, local_alignment


def get_junction(vdj, germ=False):
    global logger
    logger = log.get_logger(__name__)
    try:
        return Junction(vdj, germ)
    except Exception, err:
        logger.debug('JUNCTION ERROR: {id}'.format(id=vdj.id))
        logger.debug(traceback.format_exc())


class Junction(object):
    """docstring for Junction"""
    def __init__(self, vdj, germ):
        self.is_germ = germ
        self.in_frame = True
        self._frame_offset = 0
        self.junction_nt_start_pos = self._find_junction_nt_start(vdj)
        self.junction_nt_end_pos = self._find_junction_nt_end(vdj)
        if any([self.junction_nt_start_pos is None, self.junction_nt_end_pos is None]):
            logger.debug('JUNCTION IDENTIFICATION ERROR: {}'.format(vdj.id))
            logger.debug('RAW INPUT: {}'.format(vdj.raw_input))
            return None
        self.junction_nt = self._get_junction_nt_sequence(vdj)
        if len(self.junction_nt) % 3 > 0:
            self.in_frame = False
            self._frame_offset = 3 - len(self.junction_nt) % 3
            n_end = vdj.n2_end if vdj.d is not None else vdj.n_end
            n_end = n_end - self.junction_nt_start_pos
            ns = 'N' * self._frame_offset
            self.junction_nt = self.junction_nt[:n_end] + ns + self.junction_nt[n_end:]
        self.junction_aa = str(Seq(self.junction_nt.replace('-', ''), generic_dna).translate())
        self.cdr3_aa = self.junction_aa[1:-1]
        self.cdr3_nt = self.junction_nt[3:-3]
        if vdj.d:
            self.n1_nt = vdj.vdj_nt[vdj.n1_start:vdj.n1_end]
            self.d_nt = vdj.vdj_nt[vdj.d_start:vdj.d_end]
            self.n2_nt = vdj.vdj_nt[vdj.n2_start:vdj.n2_end]
            self.n_nt = None
            self.d_start_position_nt = self._get_d_start_position_nt(vdj)
            self.d_end_position_nt = self._get_d_end_position(vdj)
            self.d_dist_from_cdr3_start_nt = self.d_start_position_nt
            self.d_dist_from_cdr3_end_nt = self._get_d_dist_from_cdr3_end_nt(vdj)
        else:
            self.n1_nt = None
            self.d_nt = None
            self.n2_nt = None
            self.n_nt = vdj.vdj_nt[vdj.n_start:vdj.n_end]


    def _find_junction_nt_start(self, vdj):
        fr3 = vdj.v.regions.nt_seqs['FR3'][:-3]
        aln = local_alignment(fr3, vdj.vdj_nt)
        if aln:
            return aln.target_end + 1


    def _find_junction_nt_end(self, vdj):
        fr4 = vdj.j.regions.nt_seqs['FR4'][3:]
        aln = local_alignment(fr4, vdj.vdj_nt)
        if aln:
            return aln.target_begin


    def _get_junction_nt_sequence(self, vdj):
        start = self.junction_nt_start_pos
        end = self.junction_nt_end_pos
        if self.is_germ:
            return vdj.vdj_germ_nt[start:end]
        return vdj.vdj_nt[start:end]


    def _get_d_start_position_nt(self, vdj):
        a = local_alignment(self.d_nt, self.cdr3_nt,
                            gap_open_penalty=22, gap_extend_penalty=1)
        d_start = a.target_begin
        return d_start


    def _get_d_end_position(self, vdj):
        return self.d_start_position_nt + len(vdj.d.sequence)


    def _get_d_dist_from_cdr3_end_nt(self, vdj):
        return len(self.cdr3_nt) - (self.d_dist_from_cdr3_start_nt + len(self.d_nt))



#!/usr/bin/env python
# filename: abstar.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function, unicode_literals

from argparse import ArgumentParser
from glob import glob
from multiprocessing import Pool
from subprocess import Popen, PIPE

# External
from Bio import SeqIO
from abtools import log
from abtools.sequence import Sequence

import gzip
import logging
import os
import re
import sys
import time
import traceback
import warnings


#####################################################################
#
#                             ARGUMENTS
#
#####################################################################


def parse_arguments(print_help=False):
    parser = ArgumentParser("Performs germline assignment and other relevant annotation on antibody sequence data from NGS platforms.")
    parser.add_argument('-p', '--project', dest='data_dir', default=None,
                        help="The data directory, where files will be downloaded (or have previously \
                        been download), temp files will be stored, and output files will be \
                        written. During StarCluster configuration, all ephemeral drives on the master \
                        node will be combined into a single RAID volume and mounted at the data directory. \
                        Not necessary if '-o', '-t' and '-i' are provided. \
                        Default is '/data'.")
    parser.add_argument('-i', '--in', dest='input', default=None,
                        help="The input file or directory, to be split and processed in parallel. \
                        If a directory is given, all files in the directory will be iteratively processed. \
                        Required.")
    parser.add_argument('-o', '--out', dest='output', default=None,
                        help="The output directory, into which the JSON-formatted output files will be deposited. \
                        If the directory does not exist, it will be created. \
                        Required.")
    parser.add_argument('-l', '--log', dest='log', default=None,
                        help="The log file, to which log info will be written. \
                        Default is <output_directory>/abstar.log if '-o/--out' is specificied. \
                        If '-o/--out' isn't provided, default is <data_directory>/abstar.log")
    parser.add_argument('-t', '--temp', dest='temp', default=None,
                        help="The directory in which temp files will be stored. If the directory doesn't exist, \
                        it will be created. Required.")
    parser.add_argument('--sequences', dest='sequences', default=None,
                        help="Only used when passing sequences directly through the API.")
    parser.add_argument('-k', '--chunksize', dest='chunksize', default=250, type=int,
                        help="Approximate number of sequences in each distributed job. \
                        Defaults to 250. \
                        Set to 0 if you want file splitting to be turned off \
                        Don't change unless you know what you're doing.")
    parser.add_argument('-T', '--output_type', dest="output_type", choices=['json', 'imgt', 'hadoop'], default='json',
                        help="Select the output type. Options are 'json', 'imgt' and 'impala'. \
                        IMGT output mimics the Summary table produced by IMGT High-V/Quest, \
                        to maintain some level of compatibility with existing IMGT-based pipelines. \
                        JSON output is much more detailed. \
                        Hadoop output is columnar and easily converted to binary HDFS-friendly formats \
                        (Parquet, Avro) for use in Impala or other Hadoop query engines (Pig, Hive, Spark). \
                        Defaults to JSON output.")
    parser.add_argument('-m', '--merge', dest="merge", action='store_true', default=False,
                        help="Use if the input files are paired-end FASTQs \
                        (either gzip compressed or uncompressed) from Illumina platforms. \
                        Prior to running the germline assignment pipeline, \
                        paired reads will be merged with PANDAseq.")
    parser.add_argument('-P', '--pandaseq_algo', dest="pandaseq_algo", default='simple_bayesian',
                        choices=['simple_bayesian', 'ea_util', 'flash', 'pear', 'rdp_mle', 'stitch', 'uparse'],
                        help="Define merging algorithm to be used by PANDAseq.\
                        Options are 'simple_bayesian', 'ea_util', 'flash', 'pear', 'rdp_mle', 'stitch', or 'uparse'.\
                        Default is 'simple_bayesian', which is the default PANDAseq algorithm.")
    parser.add_argument('-n', '--nextseq', dest="nextseq", action='store_true', default=False,
                        help="Use if the run was performed on a NextSeq sequencer.")
    parser.add_argument('-u', '--umid', dest="uaid", type=int, default=0,
                        help="Length of the unique antibody identifiers (UAIDs) \
                        used when preparing samples, if used. \
                        Default is unbarcoded (UAID length of 0).")
    parser.add_argument('-I', '--isotype', dest="isotype", action='store_false', default=True,
                        help="If set, the isotype will not be determined for heavy chains.\
                        If not set, isotyping sequences for the appropriate species will be used.")
    parser.add_argument('-b', '--basespace', dest="basespace", default=False, action='store_true',
                        help="Use if files should be downloaded directly from BaseSpace. \
                        Files will be downloaded into the input directory.")
    parser.add_argument('-c', '--cluster', dest="cluster", default=False, action='store_true',
                        help="Use if performing computation on a Celery cluster. \
                        If set, input files will be split into many subfiles and passed \
                        to a Celery queue. If not set, input files will still be split, but \
                        will be distributed to local processors using multiprocessing.")
    parser.add_argument('-S', '--starcluster', dest="starcluster", default=False, action='store_true',
                        help="Use if performing analysis on a StarCluster instance. \
                        If set, the cluster will be configured to NFS share all ephemeral drives \
                        on the master node and Celery workers will be started on all worker nodes. \
                        Configuration only needs to be run once per cluster, so additional runs on\
                        an already-configured cluster should be run without this option.")
    parser.add_argument('-D', '--debug', dest="debug", action='count', default=0,
                        help="If set, logs additional debug information. \
                        Use -DD to print verbose exception information to screen in addition to writing to log.")
    parser.add_argument('-s', '--species', dest='species', default='human',
                        choices=['human', 'macaque', 'mouse', 'rabbit', 'b12mouse', 'vrc01mouse', '9114mouse'])
    parser.add_argument('-z', '--gzip', dest='gzip', default=False, action='store_true',
                        help="Compress the output to a gzipped file")
    parser.add_argument('--pretty', dest='pretty', default=False, action='store_true',
                        help='Pretty format json file')
    parser.add_argument('--no-padding', dest='padding', default=True, action='store_false',
                        help="If passed, will eliminate padding from json file. \
                        Don't use if you don't know what you are doing")
    if print_help:
        parser.print_help()
    else:
        args = parser.parse_args()
        return args


class Args(object):
    def __init__(self, project_dir=None, input=None, output=None, log=None, temp=None,
                 sequences=None, chunksize=250, output_type='json',
                 merge=False, pandaseq_algo='simple_bayesian',
                 nextseq=False, umid=0, isotype=False,
                 basespace=False, cluster=False, starcluster=False,
                 debug=False, print_debug=False, species='human', gzip=False):
        super(Args, self).__init__()
        self.sequences = sequences
        self.data_dir = str(project_dir) if project_dir is not None else project_dir
        self.input = str(input) if input is not None else input
        self.output = str(output) if output is not None else output
        self.log = str(log) if log is not None else log
        self.temp = str(temp) if temp is not None else temp
        self.chunksize = int(chunksize)
        self.output_type = str(output_type)
        self.merge = True if basespace else merge
        self.pandaseq_algo = str(pandaseq_algo)
        self.nextseq = nextseq
        self.uaid = int(umid)
        self.isotype = isotype
        self.basespace = basespace
        self.cluster = cluster
        self.starcluster = starcluster
        self.debug = 1 if debug else 0
        self.gzip = gzip
        if print_debug and self.debug > 0:
            self.debug == 2
        self.species = species


def validate_args(args):
    if not any([args.data_dir,
                args.sequences,
                all([args.input, args.output, args.temp])]):
        parse_arguments(print_help=True)
        sys.exit(1)
    if args.sequences:
        args.output_type = 'json'
    if all([args.sequences is not None, args.temp is None]):
        args.temp = './tmp'


#####################################################################
#
#                        FILES AND DIRECTORIES
#
#####################################################################


def make_directories(args):
    full_paths = []
    if args.data_dir and not os.path.exists(args.data_dir):
        _make_direc(args.data_dir, args)
    indir = args.input if args.input else os.path.join(args.data_dir, 'input')
    outdir = args.output if args.output else os.path.join(args.data_dir, 'output')
    tempdir = args.temp if args.temp else os.path.join(args.data_dir, 'temp')
    for d in [indir, outdir, tempdir]:
        d = os.path.abspath(d)
        full_paths.append(d)
        _make_direc(d, args)
    return full_paths


def _make_direc(d, args):
    if not os.path.exists(d):
        os.makedirs(d)
    if args.cluster:
        cmd = 'sudo chmod 777 {}'.format(d)
        p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()


def make_merge_dir(args):
    merge_dir = os.path.join(args.data_dir, 'merged')
    _make_direc(merge_dir, args)
    return merge_dir


def setup_logging(args):
    log_dir = args.output if args.output else args.data_dir
    logfile = args.log if args.log else os.path.join(log_dir, 'abstar.log')
    debug = True if args.debug > 0 else False
    print_debug = True if args.debug == 2 else False
    log.setup_logging(logfile, debug=args.debug)
    global logger
    logger = log.get_logger('abstar')


def log_options(input_dir, output_dir, temp_dir, args):
    logger.info('')
    logger.info('')
    logger.info('-' * 25)
    logger.info('ABSTAR')
    logger.info('-' * 25)
    logger.info('')
    logger.info('SPECIES: {}'.format(args.species))
    logger.info('CHUNKSIZE: {}'.format(args.chunksize))
    logger.info('OUTPUT TYPE: {}'.format(args.output_type))
    if args.merge or args.basespace:
        logger.info('PANDASEQ ALGORITHM: {}'.format(args.pandaseq_algo))
    logger.info('UAID: {}'.format(args.uaid))
    logger.info('ISOTYPE: {}'.format('yes' if args.isotype else 'no'))
    logger.info('EXECUTION: {}'.format('cluster' if args.cluster else 'local'))
    logger.info('DEBUG: {}'.format('True' if args.debug > 0 else 'False'))
    logger.debug('INPUT: {}'.format(input_dir))
    logger.debug('OUTPUT: {}'.format(output_dir))
    logger.debug('TEMP: {}'.format(temp_dir))


def list_files(d, log=False):
    if os.path.isdir(d):
        glob_pattern = os.path.join(os.path.expanduser(d), '*')
        # need to fix square brackets, or glob freaks out
        if any(['[' in glob_pattern, ']' in glob_pattern]):
            glob_pattern = re.sub(r'\[', '[[]', glob_pattern)
            glob_pattern = re.sub(r'(?<!\[)\]', '[]]', glob_pattern)
        files = sorted(glob(glob_pattern))
    else:
        files = [d, ]
    if log:
        fnames = [os.path.basename(f) for f in files]
        logger.info('FILES: {}'.format(', '.join(fnames)))
    return files


def concat_output(input_file, jsons, output_dir, args):
    bname = os.path.basename(input_file)
    if '.' in bname:
        split_name = bname.split('.')
        oprefix = '.'.join(split_name[:-1])
    else:
        oprefix = bname
    osuffix = '.json' if args.output_type == 'json' else '.txt'
    oname = oprefix + osuffix
    ofile = os.path.join(output_dir, oname)
    logger.info('Concatenating {} job outputs into a single output file.'.format(len(jsons)))
    logger.info('')
    if args.gzip:
        ohandle = gzip.open(ofile + ".gz", 'wb')
    else:
        ohandle = open(ofile, 'w')

    with ohandle as out_file:
        if args.output_type in ['json', 'hadoop']:
            for json in jsons:
                with open(json) as f:
                    for line in f:
                        out_file.write(line)
                out_file.write('\n')
        if args.output_type == 'imgt':
            for i, json in enumerate(jsons):
                with open(json) as f:
                    for j, line in enumerate(f):
                        if i == 0:
                            out_file.write(line)
                        elif j >= 1:
                            out_file.write(line)
                    out_file.write('\n')
    return ofile


def clear_temp_files(temp_files):
    for f in temp_files:
        os.unlink(f)


#####################################################################
#
#                       INPUT PROCESSING
#
#####################################################################


def download_files(input_dir):
    from utils.basespace import BaseSpace
    bs = BaseSpace()
    logger.info('')
    logger.info('BASESPACE PROJECT NAME: {}'.format(bs.project_name))
    logger.info('BASESPACE PROJECT ID: {}'.format(bs.project_id))
    num_files = bs.download(input_dir)
    logger.info('')
    logger.info('BASESPACE FILE DOWNLOAD COUNT: {}'.format(num_files))


def merge_reads(input_dir, args):
    from utils import pandaseq
    merge_dir = make_merge_dir(args)
    pandaseq.run(input_dir,
                 merge_dir,
                 args.pandaseq_algo,
                 args.nextseq)
    return merge_dir


def format_check(input_list):
    formats = []
    for f in input_list:
        formats.append(_get_format(f))
    return formats


def _get_format(in_file):
    with open(in_file) as f:
        line = f.next()
        while line.strip() == '':
            line = f.next()
        if line.lstrip().startswith('>'):
            return 'fasta'
        elif line.lstrip().startswith('@'):
            return 'fastq'
        else:
            return None


def split_file(f, fmt, temp_dir, args):
    file_counter = 0
    seq_counter = 0
    total_seq_counter = 0
    subfiles = []
    fastas = []
    if '.' in os.path.basename(f):
        out_prefix = '.'.join(os.path.basename(f).split('.')[:-1])
    else:
        out_prefix = os.path.basename(f)
    if args.chunksize != 0:
        try:
            for seq in SeqIO.parse(open(f, 'r'), fmt.lower()):
                fastas.append('>{}\n{}'.format(seq.id, str(seq.seq)))
                seq_counter += 1
                total_seq_counter += 1
                if seq_counter == args.chunksize:
                    out_file = os.path.join(temp_dir, '{}_{}'.format(out_prefix, file_counter))
                    ohandle = open(out_file, 'w')
                    ohandle.write('\n'.join(fastas))
                    ohandle.close()
                    fastas = []
                    seq_counter = 0
                    file_counter += 1
                    subfiles.append(out_file)
        except ValueError:
            print('')
            print('ERROR: invalid file.')
            print('{} is not properly formatted'.format(f))
            print(traceback.format_exception_only)
            clear_temp_files(subfiles)
            sys.exit(1)

        # We don't want our files split
    else:
        for seq in SeqIO.parse(open(f, 'r'), fmt.lower()):
            total_seq_counter += 1
        subfiles.append(f)
        file_counter = 1

    # unless the input file is an exact multiple of args.chunksize,
    # need to write the last few sequences to a split file.
    if seq_counter:
        file_counter += 1
        out_file = os.path.join(temp_dir, '{}_{}'.format(out_prefix, file_counter))
        open(out_file, 'w').write('\n' + '\n'.join(fastas))
        subfiles.append(out_file)
    logger.info('SEQUENCES: {}'.format(total_seq_counter))
    logger.info('JOBS: {}'.format(file_counter))
    return subfiles, total_seq_counter


#####################################################################
#
#                            PRINTING
#
#####################################################################


def print_input_file_info(f, fmt):
    fname = os.path.basename(f)
    logger.info('')
    logger.info('')
    logger.info(fname)
    logger.info('-' * len(fname))
    logger.info('FORMAT: {}'.format(fmt.lower()))


def print_job_stats(total_seqs, good_seqs, start_time, end_time):
    run_time = end_time - start_time
    logger.info('{} sequences contained an identifiable rearrangement'.format(good_seqs))
    logger.info('AbStar completed in {} seconds'.format(run_time))


def update_progress(finished, jobs, failed=None):
    pct = int(100. * finished / jobs)
    ticks = pct / 2
    spaces = 50 - ticks
    if failed:
        prog_bar = '\r({}/{}) |{}{}|  {}% ({}, {})'.format(finished, jobs, '|' * ticks, ' ' * spaces, pct, finished - failed, failed)
    else:
        prog_bar = '\r({}/{}) |{}{}|  {}%'.format(finished, jobs, '|' * ticks, ' ' * spaces, pct)
    sys.stdout.write(prog_bar)
    sys.stdout.flush()


#####################################################################
#
#                             JOBS
#
#####################################################################


def run_jobs(files, output_dir, args):
    sys.stdout.write('\nRunning VDJ...\n')
    if args.cluster:
        return _run_jobs_via_celery(files, output_dir, args)
    elif args.debug or args.chunksize == 0:
        return _run_jobs_singlethreaded(files, output_dir, args)
    else:
        return _run_jobs_via_multiprocessing(files, output_dir, args)


def _run_jobs_singlethreaded(files, output_dir, args):
    from utils.vdj import run as run_vdj
    results = []
    for i, f in enumerate(files):
        try:
            results.append(run_vdj(f, output_dir, args))
            update_progress(i + 1, len(files))
        except:
            logger.debug('FILE-LEVEL EXCEPTION: {}'.format(f))
            logging.debug(traceback.format_exc())
    logger.info('')
    return results


def _run_jobs_via_multiprocessing(files, output_dir, args):
    from utils.vdj import run as run_vdj
    p = Pool(maxtasksperchild=50)
    async_results = []
    for f in files:
        async_results.append((f, p.apply_async(run_vdj, (f,
                                                         output_dir,
                                                         args))))
    monitor_mp_jobs([ar[1] for ar in async_results])
    results = []
    for a in async_results:
        try:
            results.append(a[1].get())
        except:
            logger.debug('FILE-LEVEL EXCEPTION: {}'.format(a[0]))
            if args.debug:
                traceback.print_exc()
            logging.debug(''.join(traceback.format_exc()))
            continue
    p.close()
    p.join()
    return results


def monitor_mp_jobs(results):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        ready = [ar for ar in results if ar.ready()]
        finished = len(ready)
        update_progress(finished, jobs)
    sys.stdout.write('\n\n')


def _run_jobs_via_celery(files, output_dir, args):
    from utils.vdj import run as run_vdj
    async_results = []
    for f in files:
        async_results.append(run_vdj.delay(f,
                                           output_dir,
                                           args))
    succeeded, failed = monitor_celery_jobs(async_results)
    failed_files = [f for i, f in enumerate(files) if async_results[i].failed()]
    for ff in failed_files:
        logger.debug('FAILED FILE: {}'.format(f))
    return [s.get() for s in succeeded]


def monitor_celery_jobs(results):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        succeeded = [ar for ar in results if ar.successful()]
        failed = [ar for ar in results if ar.failed()]
        finished = len(succeeded) + len(failed)
        update_progress(finished, jobs, failed=len(failed))
    sys.stdout.write('\n\n')
    return succeeded, failed


def run(*args, **kwargs):
    '''
    Runs AbStar.

    Input sequences can be provided in several different formats:

        1) individual sequences as positional arguments: ``run(seq1, seq2, temp=temp, output=output)``
        2) a list of sequences, as an argument: ``run([seq1, seq2], temp=temp, output=output)``
        3) a single FASTA/Q-formatted input file, passed via ``input``
        4) a directory of FASTA/Q-formatted files, passed via ``input``

    When passing sequences (not FASTA/Q files), the sequences can be in any format recognized
    by ``abtools.sequence.Sequence``, including:

        - a raw nucleotide sequence, as a string (a random sequence ID will be assigned)
        - a list/tuple of the format ``[sequence_id, sequence]``
        - a BioPython SeqRecord object
        - an AbTools Sequence object

    Either sequences, ``project_dir``, or all of ``input``, ``output`` and ``temp`` are required.


    Examples:

        If processing a single sequence, you can pass the raw sequence, as a string::

            import abstar

            result = abstar.run('ATGC')

        or a list/tuple of the format ``[sequence_id, sequence]``::

            result = abstar.run(['seq1', 'ATGC'])

        If you pass just the raw sequence, a random sequence ID will be generated with ``uuid.uuid4()``.
        In either case, when given a single sequence, ``abstar.run()`` will return a single AbTools ``Sequence``
        object. If running multiple sequences, you can either pass each sequence as a positional argument::

            result_list = run(['seq1', 'ATGC'], ['seq2', 'CGTA'])

        or you can pass a list of sequences as the first argument, in this case using sequences parsed from a
        FASTA file using Biopython::

            from Bio import SeqIO

            fasta = open('my_sequences.fasta', 'r')
            seqs = [s for s in SeqIO.parse(fasta, 'fasta')]
            result_list = abstar.run(seqs)

        When given multiple sequences, ``abstar.run()`` will return a list of AbTools ``Sequence`` objects,
        one per input sequence.

        If you'd prefer not to parse the FASTQ/A file into a list (for example, if the input file is
        extremely large), you can pass the input file path directly, along with a temp directory and output
        directory::

            result_files = abstar.run(input='/path/to/my_sequences.fasta',
                                      temp='/path/to/temp',
                                      output='/path/to/output')

        Given a file path, ``abstar.run()`` returns a list of output file paths. In the above case,
        ``result_files`` will be a list containing a single output file path:
        ``/path/to/output/my_sequences.json``.

        If you have a directory containing multiple FASTQ/A files, you can pass the directory path
        using ``input``::

            result_files = abstar.run(input='/path/to/input',
                                      temp='/path/to/temp',
                                      output='/path/to/output')

        As before, ``result_files`` will contain a list of output file paths.

        If your input directory contains paired FASTQ files (gzip compressed or uncompressed)
        that need to be merged prior to processing with AbStar::

            result_files = abstar.run(input='/path/to/input',
                                      temp='/path/to/temp',
                                      output='/path/to/output',
                                      merge=True)

        The paired read files in ``input`` will be merged with PANDAseq prior to processing with AbStar.
        By default, PANDAseq's 'simple bayesian' read merging algorithm is used, although alternate
        algorithms can be selected with ``pandaseq_algo``.

        AbStar also provides an alternate CSV-formatted output type that mimics the `IMGT Summary file`_.
        This option is provided to minimize the effort needed to convert existing
        IMGT-based pipelines to AbStar. Alternate output is only available when passing an input file or
        directory; passing individual sequences or a list of sequences will always return Sequence objects.
        To produce IMGT-formatted output::

            result_files = abstar.run(input='/path/to/input',
                                      temp='/path/to/temp',
                                      output='/path/to/output',
                                      output_type='imgt')

        .. _IMGT Summary file: http://www.imgt.org/IMGT_vquest/share/textes/imgtvquest.html#Esummary


    Args:

        project_dir (str): Path to the project directory. Most useful when directly downloading
            files from BaseSpace, and all subdirectories will be created by AbStar.

        input (str): Path to input directory, containing FASTA/Q files. If performing
            read merging with PANDAseq, paired FASTQ files may be gzip compressed.

        output (str): Path to output directory.

        temp (str): Path to temp directory, where intermediate job files will be stored.

        log (str): Path to log file. If not provided and ``project_dir`` is provided,
            the log will be written to ``/path/to/project_dir/abstar.log``. If output is
            provided, log will be written to ``/path/to/output/abstar.log``.

        species (str): Species of the antibody sequences. Choices are 'human', 'macaque',
            'mouse' and 'rabbit'. Default is 'human'.

        isotype (bool): If True, the isotype will infered by aligning the sequence region
            downstream of the J-gene. If False, the isotype will not be determined.
            Default is True.

        umid (int): Length (in nucleotides) of the Unique Molecular ID used to barcode input RNA.
            A positive integer results in the UMID being parsed from the start of the read (or merged
            read), a negative integer results in parsing from the end of the read. Default is 0,
            which results in no UMID parsing.

        gzip (bool): If True, compresses output files with gzip. Default is False.

        pretty (bool): If True, formats JSON output files to be more human-readable. If False,
            JSON output files contain one record per line. Default is False.

        output_type (str): Options are 'json' or 'imgt'. IMGT output mimics the Summary
            table produced by IMGT High-V/Quest, to maintain a level of compatibility with
            existing IMGT-based pipelines. JSON output is much more detailed. Default is 'json'.

        merge (bool): If True, input must be paired-read FASTA files (gzip compressed or uncompressed)
            which will be merged with PANDAseq prior to processing with AbStar. If ``basespace`` is True,
            ``merge`` is automatically set to True. Default is False.

        pandaseq_algo (str): Define merging algorithm to be used by PANDAseq. Options are
            'simple_bayesian', 'ea_util', 'flash', 'pear', 'rdp_mle', 'stitch', or 'uparse'. Default is
            'simple_bayesian', which is the default PANDAseq algorithm.

        debug (bool): If ``True``, ``abstar.run()`` runs in single-threaded mode, the log is much more verbose,
            and temporary files are not removed. Default is ``False``.


    Returns:

        If the input is a single sequence, ``run`` returns a single AbTools ``Sequence`` object.

        If the input is a list of sequences, ``run`` returns a list of AbTools ``Sequence`` objects.

        If the input is a file or a directory of files, ``run`` returns a list of output files.
    '''

    warnings.filterwarnings("ignore")
    if len(args) == 1:
        # if there's a single arg, need to check if it's a single sequence...
        try:
            sequences = [Sequence(args[0]), ]
        except:
            # ...or a list of sequences
            try:
                sequences = [Sequence(s) for s in args[0]]
            except:
                print('ERROR: invalid format for sequence input:')
                for a in args:
                    print(a)
                sys.exit(1)
    # if multiple args, assume each is a sequence
    elif len(args) > 1:
        try:
            sequences = [Sequence(s) for s in args]
        except:
            print('ERROR: invalid format for sequence input:')
            for a in args:
                print(a)
            sys.exit(1)
    kwargs['sequences'] = sequences
    args = Args(**kwargs)
    validate_args(args)
    global logger
    logger = log.get_logger('abstar')
    output = main(args)
    if args.sequences is not None:
        output = [Sequence(o) for o in output]
        if len(output) == 1:
            return output[0]
    return output


def run_standalone(args):
    output_dir = main(args)


def main(args):
    if args.sequences is not None:
        from utils import output, vdj
        vdjs = vdj.process_sequences(args.sequences, args)
        return output.as_dict(vdjs)
    else:
        input_dir, output_dir, temp_dir = make_directories(args)
        setup_logging(args)
        log_options(input_dir, output_dir, temp_dir, args)
        if args.basespace:
            args.merge = True
            download_files(input_dir)
        if args.merge:
            input_dir = merge_reads(input_dir, args)
        if args.isotype:
            args.isotype = args.species
        input_files = [f for f in list_files(input_dir, log=True) if os.stat(f).st_size > 0]
        output_files = []
        for f, fmt in zip(input_files, format_check(input_files)):
            # skip the non-FASTA/Q files
            if fmt is None:
                continue
            start_time = time.time()
            print_input_file_info(f, fmt)
            subfiles, seq_count = split_file(f, fmt, temp_dir, args)
            run_info = run_jobs(subfiles, temp_dir, args)
            temp_json_files = [r[0] for r in run_info if r is not None]
            processed_seq_counts = [r[1] for r in run_info if r is not None]
            vdj_end_time = time.time()
            output_file = concat_output(f, temp_json_files, output_dir, args)
            output_files.append(output_file)
            if not args.debug:
                clear_temp_files(subfiles + temp_json_files)
            print_job_stats(seq_count, sum(processed_seq_counts), start_time, vdj_end_time)
        return output_files


if __name__ == '__main__':
    warnings.filterwarnings("ignore")
    args = parse_arguments()
    output_dir = main(args)
    sys.stdout.write('\n\n')



#!/usr/bin/python
# filename: pandaseq.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import sys
import glob
import subprocess as sp
from multiprocessing import cpu_count

from abtools import log

logger = log.get_logger('basespace')


def list_files(d):
    return sorted([f for f in glob.glob(d + '/*') if os.path.isfile(f)])


def pair_files(files, nextseq):
    pairs = {}
    for f in files:
        if nextseq:
            f_prefix = '_'.join(os.path.basename(f).split('_')[:3])
        else:
            f_prefix = '_'.join(os.path.basename(f).split('_')[:2])
        if f_prefix in pairs:
            pairs[f_prefix].append(f)
        else:
            pairs[f_prefix] = [f, ]
    return pairs


def batch_pandaseq(f, r, o, algo):
    cmd = 'pandaseq -f {0} -r {1} -A {2} -d rbfkms -T {3} -w {4}'.format(f, r, algo, cpu_count(), o)
    sp.Popen(cmd, shell=True, stderr=sp.STDOUT, stdout=sp.PIPE).communicate()


def merge_reads(files, output, algo, nextseq, i):
    files.sort()
    f = files[0]
    r = files[1]
    if nextseq:
        lane = os.path.basename(f).split('_')[-3]
        sample_id = os.path.basename(f).split('_')[0]
        sample = sample_id + '_' + lane
    else:
        sample = os.path.basename(f).split('_')[0]
    print_sample_info(i, sample)
    o = os.path.join(output, '{}.fasta'.format(sample))
    batch_pandaseq(f, r, o, algo)
    return o


def print_start_info():
    logger.info('')
    logger.info('')
    logger.info('========================================')
    logger.info('Merging reads with PANDAseq')
    logger.info('========================================')
    logger.info('')


def print_input_info(files):
    logger.info('The input directory contains {} pair(s) of files to be merged.\n'.format(len(files) / 2))


def print_sample_info(i, sample):
    logger.info('[ {} ]  Processing sample {}'.format(str(i + 1), sample))


def print_sample_end():
    logger.info('Done.')


def run(input, output, algorithm='simple_bayesian', nextseq=False):
    '''
    Merge paired-end FASTQ files with PANDAseq.

    Examples:

        To merge a directory of raw (gzip compressed) files from a MiSeq run::

            merged_files = run('/path/to/input', '/path/to/output')

        Same as above, but using the Pear_ read merging algorithm::

            merged_files = run('/path/to/input', '/path/to/output', algorithm='pear')

        To merge a list of file pairs::

            file_pairs = [(sample1_R1.fastq, sample1_R2.fastq),
                          (sample2_R1.fastq.gz, sample2_R2.fastq.gz),
                          (sample3_R1.fastq, sample3_R2.fastq)]
            merged_files = run(file_pairs, '/path/to/output')

        .. _Pear: http://sco.h-its.org/exelixis/web/software/pear/


    Args:

        input (str, list): Input can be one of three things:

                1. path to a directory of paired FASTQ files
                2. a list of paired FASTQ files
                3. a list of read pairs, with each read pair being a list/tuple
                   containing paths to two paired read files

            Regardless of what input type is provided, paired FASTQ files can be either
            gzip compressed or uncompressed.

            When providing a list of files or a directory of files, it is assumed that
            all files follow Illumina naming conventions. If your file names aren't
            Illumina-like, submit your files as a list of read pairs to ensure that
            the proper pairs of files are merged.

        output (str): Path to an output directory, into which merged FASTQ files will be deposited.
            To determine the filename for the merged file, the R1 file (or the first file in the read
            pair) is split at the first occurance of the '_' character. Therefore, the read pair
            ``['my-sequences_R1.fastq', 'my-sequences_R2.fastq']`` would be merged into ``my-sequences.fasta``.

        algorithm (str): PANDAseq algorithm to be used for merging reads. Choices are: 'simple_bayesian',
            'ea_util', 'flash', 'pear', 'rdp_mle', 'stitch', or 'uparse'. Default is 'simple_bayesian',
            which is the default PANDAseq algorithm.

        nextseq (bool): Set to ``True`` if the sequencing data was generated on a NextSeq. Needed
            because the naming conventions for NextSeq output files differs from MiSeq output.


    Returns:

        list: a list of merged file paths
    '''
    print_start_info()
    if os.path.isdir(input):
        files = list_files(input)
        pairs = pair_files(files, nextseq)
    elif type(input) in [list, tuple]:
        if all([type(i) in [list, tuple] for i in input]) and all([len(i) == 2 for i in input]):
            files = [f for sublist in input for f in sublist]
            pairs = {n: i for n, i in zip(range(len(input)), input)}
        elif all([os.path.isfile(i) for i in input]):
            files = input
            pairs = pair_files(files, nextseq)
        else:
            err = 'ERROR: Invalid input. Input may be one of three things:\n'
            err += '  1. a directory path\n'
            err += '  2. a list of file paths\n'
            err += '  3. a list of file pairs (lists/tuples containing exactly 2 file paths)'
            raise RuntimeError(err)
    else:
        err = 'ERROR: Invalid input. Input may be one of three things:\n'
        err += '  1. a directory path\n'
        err += '  2. a list of file paths\n'
        err += '  3. a list of file pairs (lists/tuples containing exactly 2 file paths)'
        raise RuntimeError(err)
    print_input_info(files)
    merged_files = []
    for i, pair in enumerate(sorted(pairs.keys())):
        if len(pairs[pair]) == 2:
            logging.info('Merging {} and {}'.format(pairs[pair][0], pairs[pair][1]))
            mf = merge_reads(pairs[pair], output, algorithm, nextseq, i)
            merged_files.append(mf)
    return merged_files



import os

try:
    from setuptools import setup
except ImportError:
    sys.exit('ERROR: setuptools is required.\n')

try:
    from pip.req import parse_requirements
except ImportError:
    sys.exit('ERROR: pip is required.\n')


if os.environ.get('READTHEDOCS', None):
    # Set empty install_requires to get install to work on readthedocs
    install_requires = []
else:
    req_file = 'requirements.txt'
    try:
        reqs = parse_requirements(req_file, session=False)
    except TypeError:
        reqs = parse_requirements(req_file)
    install_requires = [str(r.req) for r in reqs]


config = {
    'description': 'VDJ assignment and antibody sequence annotation. Scalable from a single sequence to billions of sequences.',
    'author': 'Bryan Briney',
    'url': 'www.github.com/briney/abstar/',
    # 'download_url': 'www.github.com/briney/abstar/',
    'author_email': 'briney@scripps.edu',
    'version': '0.1.5',
    'install_requires': install_requires,
    'packages': ['abstar'],
    'scripts': ['bin/abstar',
                'bin/batch_mongoimport',
                'bin/make_basespace_credfile'],
    'name': 'abstar',
    'include_package_data': True
}

setup(**config)



#!/usr/bin/env python
# filename: indels.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import re
import traceback

from abtools import log


# ------------
#  Insertions
# ------------

def find_insertions(blast_result):
    '''
    Identifies and annotates/fixes insertions. Frameshift insertions (those with a length
    not evenly divisible by 3) will be removed. Codon-length insertions will be annotated.

    Input is a BlastResult object.

    Output is a list of insertion annotations, or None if there were no codon-length
    insertions.
    '''
    logger = log.get_logger(__name__)
    try:
        insertions = []
        o = 0
        for i in re.finditer('-+', blast_result.germline_alignment):
            s = i.start() - o
            e = i.end() - o
            l = e - s
            if l % 3 == 0 or l > 3:
                insertions.append(_annotate_insertion(blast_result, s, e))
            else:
                blast_result = _fix_frameshift_insertion(blast_result, s, e)
                o += l
        return insertions if insertions else []
    except:
        logger.debug('FIND INSERTIONS ERROR: {}, {}'.format(blast_result.id,
                                                             blast_result.input_sequence))
        logger.debug(traceback.format_exc())


def _annotate_insertion(br, s, e):
    '''
    Annotates codon-length (non-frameshift) insertions.

    Input is a BlastResult object, the starting postion of the insertion (s) and the
    ending position of the insertion (e).

    Output is a dict containing insertion start position, the insertion length, and
    the inserted sequence.
    '''
    ins_start = s + br.germline_start
    ins_length = e - s
    ins_seq = br.query_alignment[s:e]
    in_frame = 'yes' if ins_length % 3 == 0 else 'no'
    return {'pos': ins_start, 'len': ins_length, 'seq': ins_seq, 'in frame': in_frame}


def _fix_frameshift_insertion(br, s, e):
    '''
    Fixes (removes) frameshift insertions.

    Input is a BlastResult object, the starting postion of the insertion (s) and the
    ending position of the insertion (e).

    Output is a modified BlastResult object.
    '''
    br.query_alignment = br.query_alignment[:s] + br.query_alignment[e:]
    br.germline_alignment = br.germline_alignment[:s] + br.germline_alignment[e:]
    br.alignment_midline = br.alignment_midline[:s] + br.alignment_midline[e:]
    br.fs_indel_adjustment += e - s
    return br


# ------------
#  Deletions
# ------------

def find_deletions(blast_result):
    '''
    Identifies and annotates/fixes deletions. Frameshift deletions (those with a length
    not evenly divisible by 3) will be removed. Codon-length deletions will be annotated.

    Input is a BlastResult object.

    Output is a list of deletion annotations, or None if there were no codon-length
    deletions.
    '''
    logger = log.get_logger(__name__)
    try:
        deletions = []
        o = 0
        for i in re.finditer('-+', blast_result.query_alignment):
            s = i.start() - o
            e = i.end() - o
            l = e - s
            if l % 3 == 0 or l > 3:
                deletions.append(_annotate_deletion(blast_result, s, e))
            else:
                blast_result = _fix_frameshift_deletion(blast_result, s, e)
                o += l
        return deletions if deletions else []
    except:
        logger.debug('FIND DELETIONS ERROR: {}, {}'.format(blast_result.id,
                                                             blast_result.input_sequence))
        logger.debug(traceback.format_exc())


def _annotate_deletion(br, s, e):
    '''
    Annotates codon-length (non-frameshift) deletions.

    Input is a BlastResult object, the starting postion of the deletion (s) and the
    ending position of the deletion (e).

    Output is a dict containing deletion start position, the deletion length, and
    the deleted sequence.
    '''
    del_start = s + br.germline_start
    del_length = e - s
    del_seq = br.germline_alignment[s:e]
    # br.nfs_indel_adjustment -= del_length
    in_frame = 'yes' if del_length % 3 == 0 else 'no'
    return {'pos': del_start, 'len': del_length, 'seq': del_seq, 'in frame': in_frame}


def _fix_frameshift_deletion(br, s, e):
    '''
    Fixes (removes) frameshift deletions.

    Input is a BlastResult object, the starting postion of the deletion (s) and the
    ending position of the deletion (e).

    Output is a modified BlastResult object.
    '''
    br.query_alignment = br.query_alignment[:s] + br.germline_alignment[s:e] + br.query_alignment[e:]
    br.fs_indel_adjustment -= e - s
    return br



#!/usr/bin/env python
# filename: starcluster.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import sys
import time
import logging
import argparse
import subprocess as sp

import paramiko


def identify_nodes():
    nodes = []
    with open('/etc/hosts', 'r') as f:
        lines = iter(f.readlines())
        line = lines.next()
        while 'master' not in line:
            line = lines.next()
        try:
            while True:
                ip, name = line.rstrip('\n').split()
                nodes.append((name, ip))
                line = lines.next()
        except StopIteration:
            pass
    return nodes


def identify_drives():
    cmd = 'ls /dev/'
    p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    drives = ['/dev/{}'.format(s) for s in stdout.split() if s.startswith('xvda') and s != 'xvda']
    return drives


def build_data_array(data_dir, log):
    drives = identify_drives()
    raid = '10' if len(drives) == 24 else '0'
    log.write('\nBuilding a RAID{} array using {} instance storage drives on master node...'.format(raid, len(drives)))
    log.flush()
    cmd = "sudo mdadm --verbose --create /dev/md0 --level={3} --chunk=256 --raid-devices={0} {1}\
           && sudo dd if=/dev/zero of=/dev/md0 bs=512 count=1\
           && sudo pvcreate /dev/md0\
           && sudo vgcreate vg0 /dev/md0\
           && sudo lvcreate -l 100%vg -n data vg0\
           && sudo mke2fs -t ext4 -F /dev/vg0/data\
           && sudo mkdir {2}\
           && echo '/dev/vg0/data {2} ext4 defaults,auto,noatime,noexec 0 0' | sudo tee -a /etc/fstab\
           && sudo mount {2}\
           && sudo chmod 777 {2}\
           ".format(len(drives), ' '.join(drives), data_dir, raid)
    p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    volume_size = get_volume_size(data_dir)
    log.write('Done.\nArray contains {} drives, mounted at {}\n'.format(len(drives), data_dir))
    if volume_size:
        log.write('{}B of space is available.\n'.format(volume_size))


def get_volume_size(volume):
    cmd = 'df -h'
    p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    for line in stdout.split('\n')[1:]:
        device, size, used, available, percent, mountpoint = line.split()
        if mountpoint == volume:
            return available
    return None


def share_data_array(data_dir, nodes, log):
    node_names = [n[0] for n in nodes]
    add_nodes_to_export(data_dir, node_names, log)
    mount_array_on_nodes(data_dir, node_names, log)


def add_nodes_to_export(data_dir, nodes, log):
    log.write('Adding nodes to /etc/exports on master node...\n')
    for node in nodes:
        cmd = 'echo "{} {}(async,no_root_squash,no_subtree_check,rw)" | sudo tee -a /etc/exports'.format(data_dir, node)
        p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
        stdout, stderr = p.communicate()
    p = sp.Popen('sudo /etc/init.d/nfs-kernel-server start', shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()


def mount_array_on_nodes(data_dir, nodes, log):
    log.write('Mounting NFS share on each node:\n')
    cmd = "sudo mkdir {0}\
           && sudo mount master:{0} {0}\
           && sudo chmod 777 {0}".format(data_dir)
    for node in nodes:
        log.write(node + '\n')
        run_remote_cmd(node, cmd)


def run_remote_cmd(node, cmd):
        '''
        Remotely runs 'cmd' on 'node' and blocks until 'cmd' completes.
        Returns 'cmd' stdout.
        '''
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(node)
        stdin, stdout, stderr = ssh.exec_command(cmd)
        channel = stdout.channel
        while not channel.exit_status_ready():
            time.sleep(1)
        return stdout.read()


def launch_celery_workers(nodes, log):
    node_names = [n[0] for n in nodes if n[0].lower() != 'master']
    cmd = 'cd /abstar\
           && chmod 777 -R /abstar\
           && sudo -u sgeadmin celery -A utils.queue.celery worker -l info --detach'
    for node in node_names:
        try:
            log.write(node + '\n')
            run_remote_cmd(node, cmd)
        except:
            log.write('FAILED\n')
            sys.exit()



def print_start_info(log):
    log.write('\n')
    log.write('\n')
    log.write('========================================\n')
    log.write('Configuring StarCluster\n')
    log.write('========================================\n')
    log.write('\n')



def run(data_dir='/data', log=sys.stdout):
    print_start_info(log)
    log.write('Identifying all cluster nodes...')
    nodes = identify_nodes()
    log.write('{} nodes identified\n'.format(len(nodes)))
    logging.info('Cluster size: {} nodes'.format(len(nodes)))

    build_data_array(data_dir, log)

    log.write('\nSharing data array with all nodes via NFS...\n')
    share_data_array(data_dir, nodes, log)
    logging.info('NFS sharing configured for all nodes.')

    log.write('\nLaunching Celery workers on all worker nodes...\n')
    launch_celery_workers(nodes, log)
    logging.info('Celery workers started all nodes.')

    log.write('\nDone. StarCluster is now properly configured for an AbStar run.\n\n')



if __name__ == '__main__':
    # standalone args
    parser = argparse.ArgumentParser("")
    parser.add_argument('-d', '--data', dest='data_dir', default='/data', help="The location in which to mount the RAID data array. Default is '/data'.")
    parser.add_argument('-l', '--log', dest='log', default=None, help="The log file. If not provided, log info will be written to sys.stdout.")
    args = parser.parse_args()

    args.log = args.log if args.log else sys.stdout

    run(data_dir=args.data_dir, log=args.log)
